#! /usr/bin/python2.7
import sys
import os
import ntpath
import copy
import time
import socket
import svgwrite
from PyQt4.QtCore import *
from PyQt4.QtGui import *
from PyQt4.QtSvg import *

os.chdir('/usr/share/iris')
sys.path.append("extraction_utils/")
sys.path.append("bin_generator/")
sys.path.append("card_functions/")
sys.path.append("misc/")
sys.path.append("gui/")

prevpath='/tmp/currentjob.svg'

import cut_preview_generator as previewgen
import pdf_paths_extractor as pdfext
import bin_generator_essentials as bingeness
import paths_interpreter as pathsint
import basic_functions as bfun
import IrisGui
import gcode_parser as gcpar

##########################################################
# Main window
class MainIris(QMainWindow, IrisGui.Ui_MainWindow):

    def __init__(self, parent=None):
        super(MainIris, self).__init__(parent)

        # Defines main window given by IrisGui.py
        self.setupUi(self)

        self.msgBox = QMessageBox()     # General Messages send by the app
        self.errbox = QErrorMessage()   # used to show several error messages
        self.errbox.setWindowTitle("Connection error")

        self.checking = bingeness.JobPacket()      # necessary to check jobs
        self.pdfpath = ""       # path of pdf design
        self.binpath = ""       # path of bin file
        self.ncpath = ""        # path of nc file
        self.filenm = ""        # filename of the design/job
        self.jobbin = None      # binary job
        self.conninst = bfun.Connection()  # connection instance, needed for sending packets
        # self.statusgetter = Status()    # necessary to get the status

        # Slot definition for button events
        self.connect(self.runbtn, SIGNAL("clicked()"), self.begincut)
        self.connect(self.stopbtn, SIGNAL("clicked()"), self.stopcut)
        self.connect(self.genjobbtn, SIGNAL("clicked()"), self.generatejob)
        self.connect(self.runperbtn, SIGNAL("clicked()"), self.runperimeter)
        self.connect(self.pausebtn, SIGNAL("clicked()"), self.pausejob)
        self.connect(self.resumebtn, SIGNAL("clicked()"), self.continuejob)
        # self.connect(self.connectbtn, SIGNAL("clicked()"), self.connectmach)
        # self.connect(self.statusgetter, SIGNAL("statusready()"), self.showstat)

        # Actions of the menu bar
        self.actionLoad_pdf.triggered.connect(self.loadpdf)
        self.actionLoad_bin.triggered.connect(self.loadbin)
        self.actionExit.triggered.connect(self.closeEvent)
        self.actionSave_job_as_bin.triggered.connect(self.savebin)
        self.actionAbout.triggered.connect(self.about)
        self.actionDetailed_stats.triggered.connect(self.detailedstats)
        self.actionReal_prev.triggered.connect(self.genbinprev)
        self.actionImport_Gcode.triggered.connect(self.gcodeconv)

        # Where will be the preview of the job
        self.svgWidget = QSvgWidget()
        self.previewarea.addWidget(self.svgWidget)

    def closeEvent(self, event):
        """
        Finishes programs execution
        Overrides the closing method, closes the socket if it is alive
        :param event:
        :return: 0
        """
        print "Closing socket"
        self.conninst.socitalk.close()
        exit(0)

    def connectmach(self):
        """
        Begins connection with the machine
        """
        # todo: implement the connection to status socket
        # self.statusgetter.run(self.ipadd.text())
        try:
            self.conninst.beginconn(self.ipadd.text())
            print 'Connecting With Host...'
        except socket.error:
            print 'Connection failed'
            self.errbox.showMessage("Check machine connection")

    def disconnmach(self):
        """
        Ends connection with the machine
        """
        self.conninst.endconn()

    def loadpdf(self):
        """
        Gets the path for the pdf design and stores on self.pdfpath, also trims the file name self.filenm
        """
        self.pdfpath = QFileDialog.getOpenFileName(self.centralwidget, 'Open file', os.getenv('HOME'), '*.pdf')
        self.pdfpath = str(self.pdfpath.toUtf8()).decode("utf-8")

        # in case no pdf has been selected
        if self.pdfpath == '':
            return

        # sets the file name
        self.filenm = ntpath.basename(self.pdfpath)

        self.statusBar.showMessage("Design loaded: " + self.filenm + ". Ready to generate binary")

        # erase previus job
        self.clearjobdata()

        print "Ready to generate bin"

    def loadbin(self):
        """
        Reads a binary job to run
        """
        # path of binary file is selected
        self.binpath = QFileDialog.getOpenFileName(self.centralwidget, 'Open file', os.getenv('HOME'), '*.bin')
        self.binpath = str(self.binpath.toUtf8()).decode("utf-8")

        # in case no path has been selected
        if self.binpath == '':
            return

        # reads the binary and checks
        self.jobbin = open(self.binpath, 'r').read()
        self.checking = bingeness.JobPacket()
        try:
            self.checking.checkpacket(self.jobbin)
        except bingeness.BadStellarisPacket as err:
            self.errorwindowsraise("Corrupted binary", "Error: load a Retina or Iris generated binary ", err)
            return

        # displays confirmation message on status bar
        self.filenm = ntpath.basename(self.binpath)
        self.statusBar.showMessage("Binary loaded: " + self.filenm + ". Ready to run")

        # no preview to show since the preview is generated from pdf extraction
        self.rstpreview()

        # writes important parameters about the current job
        self.setgenstats()

        print "Binary loaded"

    def generatejob(self):
        """
        Generates and checks a binary stream to run
        """
        initime = time.time()  # use to calculated generation and preview time

        # reading selected power
        laserpower = float(self.userpower.text())

        try:
            # reading speed
            motorspeed = self.uservel.text()
            # generates path
            polygongenerator = pdfext.PathGenerator(self.pdfpath, float(motorspeed), laserpower)
            mypolygon = polygongenerator.generate()

        except AttributeError as err:
            self.errorwindowsraise('AttributeError', 'Error: No PDF file has been loaded', err)
            return

        except pdfext.WrongNumberOfPages as err:
            self.errorwindowsraise('WrongNumberOfPages', 'This PDF contains multiple pages.\n Please select a single '
                                                         'page PDF file and try again.', err)
        else:
            print'Loading PDF file...'
            outputfile = ntpath.basename(self.pdfpath)[:-4]



            print 'Generating Binary File...'
            # translates the path to commands and makes the packet
            executor = pathsint.PolygonExecuter(mypolygon, outputfile)
            self.jobbin = executor.execute()
            print 'Binary File Generated'

            # checks generated bin
            self.checking = bingeness.JobPacket()
            self.checking.checkpacket(self.jobbin)

            # refresh general stats about the current binary
            self.setgenstats()

            # generates preview of the job
            if self.prevcheckb.isChecked():
                preview = previewgen.PreviewGenerator(copy.deepcopy(executor.getpreviewpath()))
                preview.genshowsvg(max(self.checking.width, self.checking.height))

            # shows the generated preview
            self.svgWidget.load(prevpath)

            self.statusBar.showMessage("Design loaded: " + self.filenm + ". Ready to run")

            endtime = time.time()
            print "Job ready, took ", endtime-initime, " seconds to generate\n"

    def savebin(self):
        """
        Saves the generated binary
        """
        try:
            assert isinstance(self.jobbin, str)

        except AssertionError as err:
            self.errorwindowsraise('Error', 'No design/bin has been loaded', err)

        else:
            # saves the job
            savebinobj = QFileDialog.getSaveFileName(self, "Save bin job", os.getenv('HOME'))
            binw = open(savebinobj + ".bin", 'w')
            binw.write(self.jobbin)
            binw.close()

    def begincut(self):
        """
        Sends and runs the current binary
        """
        try:
            assert isinstance(self.jobbin, str)
            self.conninst.sendjob(self.jobbin, self.ipadd.text())
        except socket.error:
            self.checkconnwarn()
        except AssertionError as err:
            self.errorwindowsraise('Error', 'No design/bin has been loaded', err)

    def runperimeter(self):
        """
        Runs the job perimeter
        """
        try:
            assert isinstance(self.jobbin, str)
            self.conninst.runperimeter(self.checking.height, self.checking.width, self.ipadd.text())
        except socket.error:
            self.checkconnwarn()
        except AssertionError as err:
            self.errorwindowsraise('Error', 'No design/bin has been loaded', err)

    def pausejob(self):
        """
        Pauses the current job execution
        """
        try:
            self.conninst.pause()
        except socket.error:
            self.checkconnwarn()

    def continuejob(self):
        """
        Resumes current job execution
        """
        try:
            self.conninst.resume()
        except socket.error:
            self.checkconnwarn()

    def stopcut(self):
        """
        Cancels the current job execution
        """
        try:
            self.conninst.canceljob()
            print "Cancelled job"
        except socket.error:
            self.checkconnwarn()

    def setgenstats(self):
        """
        Updates general stats about the current binary
        """
        self.actualpower.setText(str(self.checking.getmaxpower())+" %")
        self.cuttingtime.setText(self.checking.getcuttime())
        self.cutheight.setText(str(self.checking.height/1000.) + " in")
        self.cutwidth.setText(str(self.checking.width/1000.) + " in")

    def clearjobdata(self):
        """
        Erases the current job, the results of checking and the preview
        """
        self.jobbin = None
        self.checking = bingeness.JobPacket()
        self.rstlabels()
        self.rstpreview()

    def rstlabels(self):
        """
        clears the stats labels of the job
        """
        # clears labels of stats
        self.actualpower.setText("-")
        self.cuttingtime.setText("-")
        self.cutheight.setText("-")
        self.cutwidth.setText("-")

    def rstpreview(self):
        """
        Rewrites the preview drawing to a blank picture
        """
        # clear preview area
        dwg = svgwrite.Drawing(prevpath)
        dwg.save()
        self.svgWidget.load(prevpath)

    def checkconnwarn(self):
        """
        Shows message warning about the machine connection
        """
        print "Check machine connection"
        self.errbox.showMessage("Check machine connection")

    def errorwindowsraise(self, title, text, error):
        """
        Shows error message, given a title with some text,
        also shows the error's description
        :param title: title of the message window
        :param text: text within the message window
        :param error: error's description
        """
        print '{}'.format(error)
        self.msgBox.setWindowTitle(title)
        self.msgBox.setText(text)
        self.msgBox.setDetailedText('{}'.format(error))
        self.msgBox.exec_()

    def detailedstats(self):
        """
        Shows a dialog with detailed info about the current job
        """
        detailswindow = Detdialog(self.checking.details(self.filenm))
        detailswindow.exec_()

    def genbinprev(self):
        """
        Generates a preview from a stream of commands and
        stores it as a .svg in the gui folder,
        made for debugging
        """
        previewgen.binpreview(self.checking.commands)

    def gcodeconv(self):
        """
        Opens a nc(gcode) file and converts to a Laser Cutter job
        and stores it in self.jobbin
        """
        # gets the .nc file path
        self.ncpath = QFileDialog.getOpenFileName(self.centralwidget, 'Open file', os.getenv('HOME'), '*.nc')
        self.ncpath = str(self.ncpath.toUtf8()).decode("utf-8")

        # in case no pdf has been selected
        if self.ncpath == '':
            return

        # erase previous job
        self.clearjobdata()

        # generates the binary and checks
        self.jobbin = gcpar.gcinterpreter(self.ncpath)
        self.checking = bingeness.JobPacket()
        self.checking.checkpacket(self.jobbin)

        # sets the file name to show on gui
        self.filenm = ntpath.basename(self.ncpath)

        # writes important parameters about the current job to the gui
        self.setgenstats()
        self.statusBar.showMessage("Design loaded: " + self.filenm + ". Job ready")
        print "Binary ready"

    @staticmethod
    def about():
        """
        Shows About dialog
        """
        aboutw = Aboutdialog()
        aboutw.exec_()

##########################################################
# Dialog with detailed info
class Detdialog(QDialog):
    def __init__(self, details2show, parent=None):
        super(Detdialog, self).__init__(parent)

        # set the properties of the details dialog
        self.setWindowTitle("Detailed stats of current job")
        self.advdetails = QLabel()
        scrollarea = QScrollArea()
        layout = QVBoxLayout()
        layout.addWidget(scrollarea)
        layout.addWidget(self.advdetails)
        self.setLayout(layout)
        self.setGeometry(100, 100, 350, 500)

        # add the details
        self.advdetails.setText(details2show)

        # allows to scroll
        scrollarea.setWidget(self.advdetails)


##########################################################
# Defines the about dialog shown on the gui
class Aboutdialog(QDialog):
    def __init__(self, parent=None):
        super(Aboutdialog, self).__init__(parent)

        # defines properties of the about dialog
        self.setFixedSize(400, 300)
        self.setWindowTitle("Iris")
        self.infoiris = QLabel()
        self.labelarcos = QLabel()
        self.image = QPixmap("gui/arcoslabrev3a.png")   # top image
        self.labelarcos.setPixmap(self.image)

        layout = QVBoxLayout()
        layout.addWidget(self.labelarcos)
        layout.addWidget(self.infoiris)

        self.setLayout(layout)
        self.labelarcos.setAlignment(Qt.AlignCenter)
        self.infoiris.setAlignment(Qt.AlignCenter)
        self.infoiris.setText("\nIf you find some bugs, have comments,\n"
                              " questions, or willing to participate,\n"
                              "please contact us at\n"
                              "irisdriverteam@gmail.com\n"
                              "\nAuthors:\n"
                              "Joshua Torres Morales\n"
                              "Alexander Marin Drobinoga\n"
                              )


##########################################################
# May be useful for getting the status of the machine
class Status(QThread):
    def __init__(self, parent=None):
        super(Status, self).__init__(parent)

    # def run(self):
    #     statusinst = bfun.Connection()
    #     statusinst.connstatus(ip)
    #
    #     while True:
    #         st = statusinst.recvstatus()
    #         self.emit(SIGNAL("statusready(list)"), st)


##########################################################
# runs the gui
app = QApplication(sys.argv)
form = MainIris()
form.show()
app.exec_()
