import math
import copy
import struct
import Conf
import cut_preview_generator as previewgen
import bin_generator_essentials as bingeness
import other as o
import numpy
######################################################################
# Functions


# converts the first three ints of the commands, to bytes
def cmd2bytestring(cmd):
    for n in range(4):
        cmd[n] = struct.pack('B', cmd[n])
    return o.arr2str(cmd)

def poscalc(x0, t, vi, a):
    return x0+vi*t+0.5*a*t**2

def velcalc(t, vi, a):
    return vi+a*t

######################################################################
# Related to the conversion of vector movements into commands

class PolygonExecuter(object):

    def __init__(self, polygon, output_file):
        self.__Polygon__ = polygon
        self.__OutputFile__ = output_file
        self.__previewpath__ = None
        self.preview = None

    def setpolygon(self, polygon):
        self.__Polygon__ = polygon

    def setpreviewpath(self):
        self.__previewpath__ = copy.deepcopy(self.getpolygon())

    def getpreviewpath(self):
        return self.__previewpath__

    def getpolygon(self):
        return self.__Polygon__

    def execute(self):
        """
        Generates the commands, makes the packet and returns it
        :return: binary job string
        :rtype: str
        """
        commands = ''

        self.preview = previewgen.PreviewGenerator(copy.deepcopy(self.getpolygon()))
        self.setpreviewpath()
        # translates vector movements into commands
        for Vector in range(len(self.__Polygon__)):
            currvector = copy.deepcopy(self.__Polygon__.pop())
            # makes the cut commands
            cut = VectorMove()
            commands += cut.interpolator([currvector.getstart(), currvector.getend()],
                                         currvector.getspeed(), currvector.getacc(), currvector.getlaserpower())

        # makes and stores the fullpacket in job
        laser_bin = bingeness.JobPacket()
        job = laser_bin.makepacket(commands)

        return job


######################################################################
# Involves the generation of the cut and non-cut commands

class VectorMove:

    def __init__(self):
        self.totalcmds = 0
        self.cmdaxis = [0, 0, 0, 0]
        self.stepsaxis = 0
        self.cmdxy = [0, 0, 0, 0]
        self.stepsxy = 0
        self.nosteps = 0
        self.cmdnomove = [0, 0, 0, 0]
        self.cutcmds = []
        self.distancex = 0
        self.distancey = 0

    def movements2cmds(self, movs):
        cmds = ""
        for [mov, power] in movs:
            byte0 = (0b01 if mov[0]>0 else 0b00) + (0b10 if mov[1]>0 else 0b00)
            cmds += struct.pack('B', byte0) + \
                    struct.pack('B', abs(mov[0])) + \
                    struct.pack('B', abs(mov[1])) + \
                    struct.pack('B', power)
        return cmds

    def interpolator2(self, cutline, velocity,acc, power):
        """
        Make the commands to achieve a straight cut or movement, given by line
        in the form [[x1,y1],[x2,y2]], with a
        given velocity and power, a velocity up to 10 000 steps/s and power
        ranging from 0 to 100, returns the commands as a string
        :param cutline: line to cut or movement to do, given in inches
        :param velocity: given as a percentage
        :param power: determines fourth byte of commands [0., 100.]
        :return: cutting commands ready to make the packet or None when encounters an error
        :rtype: str or Nonetype
        """
        # rescale of speed parameter
        velocity=100.
        velocity *= min(Conf.VectorMaxVelocity[0], Conf.VectorMaxVelocity[1]) / 100

        # rescale of power parameter
        power = int(power * 255. / 100.)
        if 255 < power:
            print("Power is above max(255)")
            return

        # calculates the movement in x and y
        difx = round((cutline[1][0] - cutline[0][0]), 3)*1000
        dify = round((cutline[1][1] - cutline[0][1]), 3)*1000

        # sets direction byte of commands
        dirbyte = 0b11
        if difx < 0:
            dirbyte &= 0b10

        if dify < 0:
            dirbyte &= 0b01
        # number of commands for simultaneous movement is defined
        self.stepsxy = int(min(abs(difx), abs(dify)))

        # command for simultaneous movement is defined
        self.cmdxy = [[numpy.sign(difx), numpy.sign(dify)], power]

        # number of commands and command for just one axis is defined
        if abs(difx) < abs(dify):
            self.cmdaxis = [[0, numpy.sign(dify)], power]
            self.stepsaxis = int(abs(dify) - self.stepsxy)
        else:
            self.cmdaxis = [[numpy.sign(difx), 0], power]
            self.stepsaxis = int(abs(difx) - self.stepsxy)

        # determines total number of commands
        self.totalcmds = self.numberofcommands(velocity, difx, dify)
        # defines the number of non moving commands
        self.nosteps = self.totalcmds - self.stepsaxis - self.stepsxy
        # defines the non moving command and is stored as string of bytes
        self.cmdnomove = [[0, 0], power]
        # ordering commands, returns a list of hex strings
        self.cutcmds = self.ordercmds([[self.cmdnomove, self.nosteps],
                                       [self.cmdxy, self.stepsxy],
                                       [self.cmdaxis, self.stepsaxis]])

        # returns the commands as string
        self.cutcmds = self.movements2cmds(self.cutcmds)
        return self.cutcmds

    def interpolator(self, cutline, speed, acc, power):

        if acc==0:
            return self.interpolatorcte(cutline, speed, acc, power)


        speed *= min(Conf.VectorMaxVelocity[0], Conf.VectorMaxVelocity[1]) / 100
        # acc = Conf.VectorAcceleration[1]
        p1=cutline[0]
        p2=cutline[1]
        power = int(power * 255. / 100.)
        if 255 < power:
            print("Power is above max(255)")
            return

        #dwg = svgwrite.Drawing('test.svg')

        mvs = []

        xi=0
        yi=0

        xf=p2[0]-p1[0]
        yf=p2[1]-p1[1]
        xf=round(1000*xf)
        yf=round(1000*yf)
        h = math.sqrt(xf**2+yf**2)

        if h==0:
            h=1

        vxf = xf*speed/h
        vyf = yf*speed/h

        x=xi
        y=yi
        vx=0
        vy=0

        xtheo = 0
        ytheo = 0
        vxtheo = 0
        vytheo = 0

        ax=xf * acc/h
        ay=yf * acc/h
        oldax=ax
        olday=ay

        t=0
        tstep=1/2000.

        # calculates deceleration point

        if xf!=0:
            xdec= xf-vxf**2/(2*ax)
        else:
            xdec=0
        if yf!=0:
            ydec= yf-vyf**2/(2*ay)
        else:
            ydec=0
        if numpy.sign(xdec)!=numpy.sign(xf):
            xdec = xf/2.
        if numpy.sign(ydec)!=numpy.sign(yf):
            ydec = yf/2.
        c = 10

        oldx = xi
        oldy = yi

        brakingx = False
        brakingy = False
        while not (xf==x and yf==y):

            #checks if the final point is reached
            if y==yf:
                vy=0
                ay=0

            if x==xf:
                vx=0
                ax=0

            #check if is time to begin to decelerate
            if abs(x) > abs(xdec) and (not brakingx):
                ax=-oldax
                brakingx = True

            if abs(y) > abs(ydec) and (not brakingy):
                ay=-olday
                brakingy = True

            # calculates theoretical position and speeds
            xtheo = poscalc(xtheo, tstep, vx, ax)
            ytheo = poscalc(ytheo, tstep, vy, ay)
            vx = velcalc(tstep, vx, ax)
            vy = velcalc(tstep, vy, ay)

            # checks if estimated position
            if abs(xtheo) > abs(xf):
                xtheo = xf
            if abs(ytheo) > abs(yf):
                ytheo = yf

            # acc/dec stage
            if abs(vx)>abs(vxf):
                ax = 0
                vx = vxf

            if abs(vy) > abs(vyf):
                ay = 0
                vy = vyf

            # calculates dif between current point and expected
            xsteps = int(xtheo-x)
            ysteps = int(ytheo-y)

            # adds the movement
            mvs += [[[xsteps, ysteps], power]]

            oldx = x
            oldy = y
            # updates position
            x += xsteps
            y += ysteps

        #    dwg.add(dwg.line((oldx*c, oldy*c),(x*c, y*c), stroke=svgwrite.rgb(1, 1, 1, '%'), stroke_width=1))

        #dwg.add(dwg.line((0, 0),(xf*c, yf*c), stroke=svgwrite.rgb(1, 1, 1, '%'), stroke_width=4))
        return self.movements2cmds(mvs)

    def interpolatorcte(self, cutline, speed, acc, power):



        p1=cutline[0]
        p2=cutline[1]
        power = int(power * 255. / 100.)
        if 255 < power:
            print("Power is above max(255)")
            return

        # decfactor = 50./speed # attenuation factor of speed
        # speed = 500.   # best speed to close curves
        # power = power * decfactor     # attenuates power given the attenuation factor
        speed *= 1000./100.
        #dwg = svgwrite.Drawing('test.svg')

        mvs = []

        xi=0
        yi=0

        xf=p2[0]-p1[0]
        yf=p2[1]-p1[1]
        xf=round(1000*xf)
        yf=round(1000*yf)
        h = math.sqrt(xf**2+yf**2)

        if h==0:
            h=1

        vx = xf*speed/h
        vy = yf*speed/h

        x=xi
        y=yi

        tstep=1/2000.

        xtheo=0
        ytheo=0

        while not (xf==x and yf==y):

            #checks if the final point is reached
            if y==yf:
                vy=0

            if x==xf:
                vx=0


            # calculates theoretical position and speeds
            xtheo = poscalc(xtheo, tstep, vx, 0)
            ytheo = poscalc(ytheo, tstep, vy, 0)


            # checks if estimated position
            if abs(xtheo) > abs(xf):
                xtheo = xf
            if abs(ytheo) > abs(yf):
                ytheo = yf



            # calculates dif between current point and expected
            xsteps = int(xtheo-x)
            ysteps = int(ytheo-y)

            # adds the movement
            mvs += [[[xsteps, ysteps], power]]
            # updates position
            x += xsteps
            y += ysteps

        #    dwg.add(dwg.line((oldx*c, oldy*c),(x*c, y*c), stroke=svgwrite.rgb(1, 1, 1, '%'), stroke_width=1))

        #dwg.add(dwg.line((0, 0),(xf*c, yf*c), stroke=svgwrite.rgb(1, 1, 1, '%'), stroke_width=4))
        return self.movements2cmds(mvs)



    @staticmethod
    def numberofcommands(velocity, deltax, deltay):
        """
        Receives the velocity parameter, and given the distance
        calculates and returns the total number of commands
        :param velocity: given in steps per second
        :param deltax: displacement on x direction, given in steps
        :param deltay: displacement on y direction, given in steps
        :return: integer, total number of commands neccesary to achieve the given velocity
        """
        # calculates distance of cut
        distance = math.sqrt(deltax ** 2 + deltay ** 2)
        # calculate the time of the cut
        timecut = distance / velocity
        # total number of commands
        return int(timecut * Conf.commandsPerSec)

    def ordercmds(self, allcmds):
        """
        Receives a list with commands with their respective number of times
        that must appear in the final cutting string, and merges close to
        an homogenous pattern
        :param allcmds: commands to merge with their respective amount o occurencies,
        must have the form [[command0, amountcommand0],[command1, amountcommand1],...]
        where commands are hex strings and amounts are integers
        :return: orderedcmds, which is a list of commands, where every command is a hex string
        :rtype: list
        """
        # keep track of how many commands had been written
        counters = [0]*len(allcmds)

        # where will be the merged commands
        orderedcmds = []

        # where will be the commands to add
        cmds = []

        # removes commands with 0 regularity, also calculates total number of commands
        total = 0
        for [cmd, cmdtimes] in allcmds:
            total += cmdtimes
            if cmdtimes != 0:
                cmds += [[cmd, cmdtimes]]
        # elaborates the ordered list of commands to return
        n = 0
        while True:
            for k in range(len(cmds)):
                if n == round(float(total)/cmds[k][1]*counters[k]) and counters[k] != cmds[k][1]:
                    orderedcmds += [cmds[k][0]]
                    counters[k] += 1
            n += 1
            # conditional to exit, if all commands had been added, then exit

            if len(orderedcmds) == total:
                break

        return orderedcmds
