import socket
import time
import struct
import math
import sys
sys.path
sys.path.append('bin_generator')
import bin_generator_essentials as bingeness
import Conf
sys.path.append('misc')
import other as o

###############################################################################
# Related to functions of italk and status port

class Connection:
    def __init__(self):
        """
        Initializes attributes
        """

        self.ipaddr = ""

        # sockets to communicate with laser card
        self.socjobs = socket.socket()
        self.socitalk = socket.socket()
        self.socstatus = socket.socket()

        self.stat = " "  # status packet
        self.mode = 0   # defines the current action of the machine
        self.homed = 2  # homed or not homed
        self.posx = 0   # x position of laser
        self.posy = 0   # y position of laser
        self.checksumint = 0    # checksum value of status packet
        self.validstat = False  # defined to True if checksum values are consistent

    def beginconn(self, ipaddress):
        """
        Begins connections with command's socket, italk,
        given the ip address and then prints italk response
        :param ipaddress: IP address of the machine, as a string
        """
        self.ipaddr = ipaddress

        print "Beginning connection"
        try:
            self.socitalk.close()
        except:
            pass

        # Connects to italk
        self.socitalk = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socitalk.getsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE)
        self.socitalk.settimeout(1)
        self.socitalk.connect((self.ipaddr, 12345))
        # print response
        print "italk: " + self.socitalk.recv(2048)

        print "Connected with machine card"

    def endconn(self):
        """
        Terminates connections with the machine and return italk response
        """

        # sends the command to disconnect and shows the response
        self.socitalk.send("bye\n")
        print "italk: " + self.socitalk.recv(2048)

        # close sockets
        self.socitalk.close()

        print "Connection finished"

    def canceljob(self):
        """
        Sends the abort command given a previusly connected socket with italk port
          then prints italk response
        """
        self.socitalk.send("abort\n")
        print "Sending cancel job command"
        print "italk: ", self.socitalk.recv(2048)

    def pause(self):
        """
        Makes the machine to pause the running job, given a previusly connected socket with italk port
          then prints italk response
        """
        self.socitalk.send("pause\n")
        print "Sending pause job execution command"
        print "italk: ", self.socitalk.recv(2048)

    def resume(self):
        """
        Makes the machine to continue the running job, given a previusly connected socket with italk port
          then prints italk response
        """
        self.socitalk.send("resume\n")
        print "Sending resume job execution command"
        print "italk: ", self.socitalk.recv(2048)

    def sendconfig(self, height=0, width=0):
        """Sends the configuration struct
        :param height: determines perimeter on y axis
        :param width: determines perimeter on x axis
        """
        # elaborates the configuration packet
        body0 = '\x05'+19*'\x00'
        cmd0 = self.struct2send(1, body0)
        jpacket = bingeness.JobPacket()
        config = jpacket.makeconfstruct(height, width)
        cmdconfig = self.struct2send(0, config)

        # sends the configuration packet
        self.cmdsender(cmd0+cmdconfig)

        print "Sending configuration structure to machine card"

    def readrunjob(self, binfile, ipadd):
        """
        Opens a binary job and then begins the machine execution
        :param binfile: name of a binary file to send, as a string
        :param ipadd: ip address of the machine, as a string
        """
        file2send = open(binfile, 'r')
        self.sendjob(file2send.read(), ipadd)

    def sendjob(self, jobbin, ip):
        """
        Accepts the job as a binary stream and then runs that job
        :param jobbin: binary string to send to italk
        :param ip: ip address of the machine, as a string
        """

        # connecting with the machine
        self.beginconn(ip)

        # first step of the protocol to send a job
        self.socitalk.send("xjob\n")
        print "italk: ", self.socitalk.recv(2048)

        # sends the size of the full packet
        self.socitalk.send("immediate " + str(len(jobbin)) + "\n")
        print "italk: ", self.socitalk.recv(2048)

        # third step of protocol
        self.socitalk.send("data\n")
        print "italk: ", self.socitalk.recv(2048)

        # connects to another port which would receive the packet
        self.socjobs = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socjobs.connect((self.ipaddr, 12346))

        # prepares the machine to receive the packet
        self.socitalk.send("sending\n")
        print "italk: ", self.socitalk.recv(2048)

        # send the packet
        self.socjobs.send(jobbin)
        self.socjobs.close()
        print "italk: ", self.socitalk.recv(2048)

        # must insert a conditional to check the retrieval of data

        # send the command to run the job
        self.socitalk.send("run\n")
        print "italk: ", self.socitalk.recv(2048)
        print "Running job"

        # included in regular protocol, but seems optional, the same with abort command
        bodypower = '\x16'+struct.pack('B', Conf.LaserPowerMax)+'\x00'*18
        cmdpower = self.struct2send(1, bodypower)
        self.cmdsender(cmdpower)

    ############################################################
    # Building commands related methods
    @staticmethod
    def struct2send(identifier, messg):
        """
        Involves the sending of structs/commands to laser(italk port),
        receives a indentifier as an int a the actual message/command
        to send
        :param identifier: identifier for the command, it's predefined, int type
        :param messg: core of the command, it's predefined, string type
        :return: formated structure
        :rtype: str
        """

        # just follows the construction of a command
        struct2send = ['\xaa', struct.pack('B', identifier)]
        struct2send += struct.pack('<H', len(messg))
        struct2send += messg
        struct2send += ['\xbb\xcc'] * 4

	print struct2send

        return o.arr2str(struct2send)

    def cmdsender(self, cmd):
        """
        Accepts the command as a string and sends the command to italk
        which includes the sending of the pass <size>\n message
        :param cmd: structure to send, as a bin stream
        """
        # necessary step of protocol
        self.socitalk.send('pass ' + str(len(cmd)) + '\n')
        print "italk: ", self.socitalk.recv(2048)
        # send the command
        self.socitalk.send(cmd)
        print "italk: ", self.socitalk.recv(2048)

    ############################################################
    # Moving laser related methods

    def jogxy(self, stepsx, stepsy):
        """
        Makes and send a command to move the laser head, where stepsX/Y
        are in steps(1000steps =1 inch) and may be positive(up and right)
        or negative(down and left), also waits t seconds necessary to execute the command
        :param stepsx: amount of steps to travel on x axis
        :param stepsy: amount of steps to travel on y axis
        """

        # makes command for jog
        jogmessg = '\x06\x00\xff\x00\x01\x00\x00\x00' + \
                   struct.pack('<i', stepsx) + \
                   struct.pack('<i', stepsy) + \
                   '\x00\x00\x00\x00'
        jogcmd = self.struct2send(1, jogmessg)  # identifier must be 1

        # sends command
        self.cmdsender(jogcmd)

        # distance
        d = math.sqrt(stepsx**2 + stepsy**2)
        # velocity
        v = 10000.
        # estimated time
        t = d/v

        time.sleep(t)

    def infmove(self, direction):
        """
        Begins the continuous movement on a given direction
        :param direction: specifies direction, of the char type, 'r'-right, 'l'-left, 'u'-upwards, 'd'-downwards
        """
        # preparation command
        bodyini = '\x0c'+19*'\x00'
        cmdini = self.struct2send(1, bodyini)
        self.cmdsender(cmdini)

        # definition of moving command
        if direction == 'r':
            dirbyte = 1
        elif direction == 'l':
            dirbyte = 2
        elif direction == 'u':
            dirbyte = 3
        elif direction == 'd':
            dirbyte = 4
        else:
            dirbyte = 1
        bodybeginm = '\x03'+struct.pack('B', dirbyte)+18*'\x00'
        cmdbeginm = self.struct2send(1, bodybeginm)

        # begins movement
        self.cmdsender(cmdbeginm)

        # allows to move the laser 1 second in the given direction
        time.sleep(1)

        # ending movement command
        bodyend = '\x73\xf7\x01\x00'
        cmdend = self.struct2send(6, bodyend)
        self.cmdsender(cmdend)

    def runperimeter(self, h, w, ipadd):
        """
        Runs job perimeter given it height and width, the ip address is also needed to send the commands
        :param h: height, given in steps as an integer
        :param w: width, given in steps as an integer
        :param ipadd: ip address as a string
        :return: None
        """
        # connecting
        self.beginconn(ipadd)

        # send config, enables running perimeter from card panel
        # self.sendconfig(h, w)

        # moves laser
        self.jogxy(w, 0)
        self.jogxy(0, -h)
        self.jogxy(-w, 0)
        self.jogxy(0, h)

        print "Running jop perimeter"

        # ending connection
        self.endconn()

    ############################################################
    # Status related methods

    def connstatus(self, ipadd):
        """
        Connects to status socket 12347
        :param ipadd: is the ip address, as a string
        """
        self.socstatus = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socstatus.connect((ipadd, 12347))
        print "Beginning status connection"

    def endconnstatus(self):
        """
        Ends connection with status socket
        """
        self.socstatus.close()
        print "Ending status connection"

    def recvstatus(self):
        """
        Receives status, parses it and returns relevant properties of machine state as a dict
        or None in case status stream is corrupted
        """
        status = self.socstatus.recv(68)
        return self.parsestatus(status)

    def getstatus(self):
        """
        Just formats an returns the status struct as a string,
        ready to print, made for testing porpuses
        :return: info about the state of the machine ready to print
        :rtype: str
        """
        stat = ""
        # stat += "\nSerial: " + str(self.serial)

        if self.mode == 0:
            action = 'Idle'
        elif self.mode == 1:
            action = 'Vectoring'
        elif self.mode == 2:
            action = 'Raster'
        else:
            action = 'Undefined'

        stat += "\nMode: " + str(self.mode) + action

        if self.homed & 1:
            ishome = 'Yes'
        else:
            ishome = 'No'

        stat += "\nHomed: " + ishome

        # stat += "\nint0: " + str(hex(self.int0))
        # stat += "\nint1: " + str(hex(self.int1))
        # stat += "\nint2: " + str(hex(self.int2))
        #
        # stat += "\nbyte0 byte1 byte2 byte3: " + \
        # str(hex(self.byte0)) + str(hex(self.byte1)) + \
        # str(hex(self.byte2)) + str(hex(self.byte3))
        #
        # stat += "\nuint1: " + str(hex(self.uint1))

        stat += "\nPosition: " + str(self.posx) + str(self.posy)
        # stat += "\nMove to: " + str(self.movex) + str(self.movey)

        stat += "\nChecksum: " + str(self.checksumint)  # given on the status packet
        stat += "\nChecksum: " + str(self.checksumstat(self.stat))  # calculated

        # stat += "\nCard hours: " + str(self.cardhours)
        # stat += "\nKey: " + str(self.keycard)
        # stat += "\nuint5: " + str(self.uint5)

        if self.validstat:
            stat += '\nValid'

        return stat

    def parsestatus(self, status):
        """
        Obtains the fields of the status struct, sets relevant properties of machine state,
        returns some of this properties if checksum int are consistent, if not returns None
        :param status: string, the struct received from the status port as a string
        :return: dictionary, with the current mode of the machine, which is an integer,
        or None if status stream is corrupted
        """
        self.stat = status
        # self.serial = struct.unpack('<I', status[0:4])[0]
        self.mode = struct.unpack('<H', status[4:6])[0]
        self.homed = struct.unpack('<H', status[6:8])[0]

        # self.int0 = struct.unpack('<i', status[8:12])[0]
        # self.int1 = struct.unpack('<i', status[12:16])[0]
        # self.int2 = struct.unpack('<i', status[16:20])[0]
        #
        # self.byte0 = struct.unpack('B', status[20])[0]
        # self.byte1 = struct.unpack('B', status[21])[0]
        # self.byte2 = struct.unpack('B', status[22])[0]
        # self.byte3 = struct.unpack('B', status[23])[0]
        #
        # self.uint1 = struct.unpack('<I', status[24:28])[0]

        self.posx = struct.unpack('<i', status[28:32])
        self.posy = struct.unpack('<i', status[32:36])

        # self.posz=struct.unpack('<i', status[36:40])
        # self.startx = struct.unpack('<i', status[40:44])[0]
        # self.starty = struct.unpack('<i', status[44:48])[0]
        # self.startz = struct.unpack('<i', status[48:52])[0]

        self.checksumint = struct.unpack('<I', status[52:56])[0]

        # self.cardhours = struct.unpack('<I', status[56:60])[0]
        # self.keycard = struct.unpack('<I', status[60:64])[0]
        # self.uint5 = struct.unpack('<I', status[64:68])[0]

        # to check the integrity of the data
        if self.checksumstat(status) == self.checksumint:
            self.validstat = True
            return {"mode": self.mode}
        else:
            return None

    @staticmethod
    def checksumstat(status):
        """
        Accepts the status packet as a binary stream and elaborates checksum of status packet
        :rtype: int
        :param status: status struct received from status port as a string
        :return: checksum integer
        """
        status = o.str2arr(status)
        for s in range(len(status)):
            status[s] = struct.unpack('B', status[s])[0]

        status[64:66] = [0, 0]
        status[52:56] = [0, 0, 0, 0]

        if len(status) == 68:
            chsum = (2**32)-1
            for i in range(len(status)):
                for j in range(0x20):
                    num = chsum & 1
                    chsum = (chsum >> 1) | ((status[i] & 1) << 0x1f)
                    status[i] >>= 1
                    if num != 0:
                        chsum ^= 0xedb88320
            return chsum+1

        return
