import cutterpath
import bin_generator_essentials as bingeness
import paths_interpreter as pathsint


###################################################################
# Converts nc to laser cutter commands

def gcinterpreter(ncfile):
    """
    Reads a ncfile and converts the commands to laser cut commands
    and generates the full packet, ready to be sent to the laser
    or written to a file
    :param ncfile: name of the nc file to read
    :return: laser cutter job as a binary stream
    """
    # reads the nc file
    nc2read = open(ncfile, 'r')
    nclines = nc2read.readlines()


    cmds=""
    oldx=0.
    oldy=0.

    # instance used to call the interpolator
    cut = pathsint.VectorMove()

    # parses the lines one by one
    for line in nclines:

        params=line.split()

        # checks if there are no parameters
        if len(params)==0:
            continue

        # movement command of GCode
        if params[0]=='G01':
            x = round(float(params[1][1:]), 3)
            y = round(float(params[2][1:]), 3)

            speed = params[3]
            speed = float(speed[1:]) * (10./6.)

            # appends commands
            cmds += cut.interpolatorcte([[oldx, oldy], [x, y]], speed, None, 0)
            oldx = x
            oldy = y

    # makes the packet
    laser_bin = bingeness.JobPacket()
    job = laser_bin.makepacket(cmds)

    return job

