import math


#################################################################
# Describes a laser movement with a given power
class VectorCut(object):
    def __init__(self, start, end, speed, power=0.0):
        """
        Initializes relevant attributes of a line to cut
        :param start: initial position in inches
        :param end: final position in inches
        :param power: laser power, ranging from 0 to 100
        :return:
        """
        self.__StartingPoint__ = start
        self.__EndingPoint__ = end
        self.__LaserPower__ = power
        self.__speed__ = speed
        self.__DeltaX__ = abs((self.__EndingPoint__[0] - self.__StartingPoint__[0]))
        self.__DeltaY__ = abs((self.__EndingPoint__[1] - self.__StartingPoint__[1]))

    # Print overload
    def __repr__(self):
        return '< {} , {} > \n'.format(self.__StartingPoint__, self.__EndingPoint__)

    def __add__(self, other):
        """
        :param other:A second Vector Class Object
        :return: This method returns a new Vector Cut formed by the starting point of the
                 first vector, the ending point of the second and the power of the first vector
        """
        assert type(other) is VectorCut, \
            'Other is not a Vector Class Object: Object type received {}'.format(type(other))
        return VectorCut(self.__StartingPoint__, other.getend(), self.__LaserPower__)

    def setstart(self, x0, y0):
        self.__StartingPoint__[0] = x0
        self.__StartingPoint__[1] = y0

    def getstart(self):
        return self.__StartingPoint__

    def setend(self, xf, yf):
        self.__EndingPoint__[0] = xf
        self.__EndingPoint__[1] = yf

    def getend(self):
        return self.__EndingPoint__

    def setlaserpower(self, power=0):
        self.__LaserPower__ = power

    def getlaserpower(self):
        return self.__LaserPower__

    def setspeed(self, speed):
        self.__speed__ = speed

    def getspeed(self):
        return self.__speed__

    def scalevector(self, scale=1.0):
        """
        Rescales the starting and ending points according to scale parameter
        :param scale: factor to scale the vectors, float type
        """
        self.__StartingPoint__[0] *= scale
        self.__StartingPoint__[1] *= scale
        self.__EndingPoint__[0] *= scale
        self.__EndingPoint__[1] *= scale

    def getmagnitude(self):
        return math.sqrt(math.pow(self.__DeltaX__, 2) + math.pow(self.__DeltaY__, 2))

    def getangle(self):
        return math.degrees(math.atan2(self.__DeltaY__, self.__DeltaX__))

    def getdeltax(self):
        return self.__DeltaX__

    def getdeltay(self):
        return self.__DeltaY__
